from datetime import datetime
from typing import List

from ws_barcode_scanner.memory_map import MemoryMap
from ws_barcode_scanner.serial_port import SerialPort


class BarcodeScanner:
    __serial_port: SerialPort
    __last_code: bytes
    __last_timestamp: datetime
    memory_map: MemoryMap
    """
    Interface to the device memory
    """

    def __init__(self, serial_port: str):
        self.__serial_port = SerialPort(serial_port)

        self.memory_map = MemoryMap(self.__serial_port)

        self.__last_code = b""
        self.__last_timestamp = datetime.now()

    def save_to_flash(self) -> None:
        """Save the current settings to the persistent flash memory"""
        self.__serial_port.save_to_flash()

    def restore_factory_settings(self) -> None:
        """Restore factory settings"""
        self.__serial_port.restore_factory_settings()

    def query_for_codes(self) -> List[bytes]:
        """Query for scanned codes"""
        response = self.__serial_port.read_all().strip()
        if response:
            codes = response.split(b"\r")
            self.__last_code = codes[-1]
            self.__last_timestamp = datetime.now()
            return codes
        return []

    @property
    def last_code(self) -> bytes:
        """The last scanned barcode"""
        return self.__last_code

    @property
    def last_timestamp(self):
        """The timestamp when the last barcode was scanned (startup time if no code was scanned yet)"""
        return self.__last_timestamp



from typing import List, Tuple, Union, overload

from ws_barcode_scanner.serial_port import SerialPort
from ws_barcode_scanner.utils import bits_to_int, int_to_bits


class MemoryMap:
    """
    Interface to the device memory
    """

    __serial_port: SerialPort

    def __init__(self, serial_port: SerialPort) -> None:
        self.__serial_port = serial_port

    @overload
    def __getitem__(self, item: Union[int, slice]) -> int:
        ...

    @overload
    def __getitem__(self, item: Tuple[Union[int, slice], int]) -> bool:
        ...

    @overload
    def __getitem__(self, item: Tuple[Union[int, slice], slice]) -> int:
        ...

    def __getitem__(self, key):
        """
        Read data from memory

        Examples
        --------

        >>> # read bytes as integers
        >>> self[0x0000]  # first byte
        202
        >>> bin(self[0x0000])  # same number as binary
        '0b11001010'
        >>> bin(self[0x0000:0x0002])  # first two bytes
        '0b1100101011001010'

        >>> # read bit as bool
        >>> self[0x0000, 0]  # first bit of first byte
        True

        >>> # read bit slices as integers
        >>> bin(self[0x0000, 0:4])  # read first four bits of first byte
        '0b1100'
        """
        if isinstance(key, (int, slice)):
            byte_index, bit_index = key, None
        elif isinstance(key, tuple) and len(key) == 2:
            byte_index, bit_index = key
        else:
            raise IndexError(f"Index not understood: {key!r}")

        # read bytes from device
        if isinstance(byte_index, int):  # self[0] -> int, 8 bits
            num_bytes = 1
            byte_data = self.__serial_port.read_address(byte_index, num_bytes)
        elif isinstance(byte_index, slice):  # self[0:2] -> int, 16 bits
            byte_index = self.__normalize_slice(byte_index)
            num_bytes = byte_index.stop - byte_index.start
            byte_data = self.__serial_port.read_address(byte_index.start, num_bytes)
        else:
            raise IndexError(f"Byte index not understood: {byte_index!r}")

        # if full bytes are requested: return them
        if bit_index is None:
            return byte_data

        byte_data_as_bits = int_to_bits(byte_data, n_bits=num_bytes * 8)

        # return requested bits
        if isinstance(bit_index, int):  # self[0, 0] -> bool
            return byte_data_as_bits[bit_index]
        elif isinstance(bit_index, slice):  # self[0, 2:4] -> int, 2 bits
            bit_index = self.__normalize_slice(bit_index)
            return bits_to_int(byte_data_as_bits[bit_index])

        raise IndexError(f"Bit index not understood: {bit_index}")

    @overload
    def __setitem__(self, key: Union[int, slice], value: Union[int, bytes]) -> None:
        ...

    @overload
    def __setitem__(self, key: Tuple[Union[int, slice], int], value: Union[int, bool]) -> None:
        ...

    @overload
    def __setitem__(self, key: Tuple[Union[int, slice], slice], value: Union[int, bytes]) -> None:
        ...

    def __setitem__(self, key, value) -> None:
        r"""
        Write data to memory

        Examples
        --------

        >>> # write bytes
        >>> self[0x0000] = 0b10100101  # write `10100101` to first byte
        >>> self[0x0000] = 165  # same
        >>> self[0x0000] = bytes.fromhex("A5")  # same
        >>> self[0x0000] = b"\xA5"  # same

        >>> # write bit
        >>> self[0x0000, 0] = 0  # set first bit of first byte to 0
        >>> self[0x0000, 0] = False  # same

        >>> # write bit slices
        >>> self[0x0000, -2:] = 0b00  # set last two bits of first byte to `00`
        >>> self[0x0000, -2:] = 0  # same
        """
        if isinstance(key, (int, slice)):
            byte_index, bit_index = key, None
        elif isinstance(key, tuple) and len(key) == 2:
            byte_index, bit_index = key
        else:
            raise TypeError(f"Index not understood: {key!r}")

        if isinstance(byte_index, int):
            address, n_bytes = byte_index, 1
        elif isinstance(byte_index, slice):
            byte_index = self.__normalize_slice(byte_index)
            address, n_bytes = key.start, key.stop - key.start
        else:
            raise IndexError(f"Byte index not understood: {byte_index!r}")

        if isinstance(value, bytes):
            int_value = int.from_bytes(value, byteorder="big")
        elif isinstance(value, int):
            int_value = value
        else:
            raise TypeError("Value must be bytes or integer")

        # write bytes if no bit index is given
        if bit_index is None:
            self.__serial_port.write_to_address(address, n_bytes, int_value)
            return

        # read bytes that have to be partially modified
        bits: List[bool] = list(int_to_bits(self[byte_index], n_bits=n_bytes * 8))

        if isinstance(bit_index, int):
            if isinstance(int_value, bool) or (isinstance(int_value, int) and int_value in (0, 1)):
                bits[bit_index] = bool(int_value)
            else:
                raise ValueError("Can only write True, False, 0, or 1 to a bit address")
        elif isinstance(bit_index, slice):
            bit_index = self.__normalize_slice(bit_index)
            bits[bit_index] = int_to_bits(int_value, n_bits=bit_index.stop - bit_index.start)
        else:
            raise IndexError(f"Bit index not understood: {bit_index}")

        self.__serial_port.write_to_address(address, n_bytes, bits_to_int(bits))

    @staticmethod
    def __normalize_slice(key: slice) -> slice:
        """
        Check and normalize the given byte slice s.t. start is an integer, stop is not None, and step is None

        Parameters
        ----------
        key
            A slice to be used as byte index

        Returns
        -------
        The normalized slice

        Raises
        ------
        IndexError
            If the step was given or stop was not given
        """
        if key.step is not None:
            raise IndexError("Slicing with step != 1 is not supported")
        if key.stop is None:
            raise IndexError("Slicing with open stop is not supported")
        if key.start is None:
            return slice(0, key.stop, None)
        return key



from contextlib import contextmanager
from threading import Lock
from typing import Iterator, Sequence, Tuple


@contextmanager
def lock_threading_lock(lock: Lock, *, timeout: float) -> Iterator[None]:
    """Context manager for using a `threading.lock` with a timeout"""
    if not timeout > 0:
        raise ValueError("Timeout must be positive")

    if not lock.acquire(timeout=timeout):
        raise TimeoutError(f"Could not acquire lock after {timeout} seconds")

    try:
        yield
    finally:
        lock.release()


def int_to_bits(num: int, n_bits: int) -> Tuple[bool, ...]:
    """
    Convert an integer to its bit representation.

    Examples
    --------

    >>> int_to_bits(3, 2)
    (True, True)
    >>> int_to_bits(0xF1, 8)
    (True, True, True, True, False, False, False, True)
    """
    max_num_representable_by_n_bits = 2**n_bits - 1
    if num > max_num_representable_by_n_bits:
        raise ValueError(f"{n_bits} can only represent numbers <= {max_num_representable_by_n_bits}, got {num}")
    return tuple(bool(num & (2 ** (n_bits - i - 1))) for i in range(n_bits))


def bits_to_int(bits: Sequence[bool]) -> int:
    return sum(bit * 2**i for (i, bit) in enumerate(bits[::-1]))



import sys

if sys.version_info[:2] < (3, 8):
    import importlib_metadata as metadata
else:
    from importlib import metadata

from .barcode_scanner import BarcodeScanner

__version__ = metadata.version("ws-barcode-scanner")

__all__ = ["__version__", "BarcodeScanner"]



from functools import lru_cache
from threading import Lock

from serial import Serial

from ws_barcode_scanner.utils import lock_threading_lock


class SerialPort:
    __serial_port: Serial
    __port_lock: Lock
    __timeout: float

    def __init__(self, com_port: str, *, timeout: float = 1) -> None:
        self.__serial_port = Serial(com_port, timeout=timeout, write_timeout=timeout)
        self.__port_lock = Lock()
        self.__timeout = timeout

    def read_all(self) -> bytes:
        """Read all bytes from the input buffer"""
        with lock_threading_lock(self.__port_lock, timeout=self.__timeout):
            return self.__serial_port.read(self.__serial_port.in_waiting)

    def read_address(self, address: int, size: int) -> int:
        """Read `size` bytes from the given memory `address`"""
        hex_address = self.__prepare_address(address)
        hex_size = self.__prepare_size(size)

        expected_response_prefix = f"020000{hex_size}"
        expected_response_length = 6 + size  # header: 2, types: 1, size: 1, crc: 2

        hex_response = self.__communicate(
            f"7E000701{hex_address}{hex_size}", f"020000{hex_size}", expected_response_length
        )

        hex_data = hex_response[len(expected_response_prefix) : -4]  # last 2 bytes (4 hex digits) are crc

        return int(hex_data, base=16)

    def write_to_address(self, address: int, size: int, data: int) -> None:
        """Write the given `size` bytes of `data` to the `address`"""
        hex_size = self.__prepare_size(size)
        hex_address = self.__prepare_address(address)
        hex_data = hex(data)[2:].zfill(2 * size)

        expected_response_prefix = bytes.fromhex("0200000100")
        expected_response_length = len(expected_response_prefix) + 2  # 2 bytes crc suffix

        self.__communicate(f"7E0008{hex_size}{hex_address}{hex_data}", "0200000100", expected_response_length)

    def save_to_flash(self) -> None:
        expected_hex_response = "02000001003331"
        self.__communicate("7E000901000000", expected_hex_response, len(expected_hex_response) // 2)

    def restore_factory_settings(self) -> None:
        expected_hex_response = "02000001003331"
        self.__communicate("7E0009010000FF", expected_hex_response, len(expected_hex_response) // 2)

    def __communicate(self, hex_message: str, expected_hex_prefix: str, expected_bytes: int) -> str:
        """Send `hex_message` to the port and expect `expected_bytes` bytes with the given `expected_hex_prefix`"""
        expected_hex_prefix = expected_hex_prefix.upper()

        message_bytes = bytes.fromhex(hex_message)
        crc_bytes = self.__compute_crc(message_bytes[2:]).to_bytes(length=2, byteorder="big")

        with lock_threading_lock(self.__port_lock, timeout=self.__timeout):
            self.__serial_port.write(message_bytes + crc_bytes)
            raw_response = self.__serial_port.read(expected_bytes)

        hex_response = raw_response.hex().upper()

        if len(raw_response) != expected_bytes:
            raise TimeoutError(f"Expected to receive {expected_bytes} bytes, got '{hex_response}'")

        if not hex_response.startswith(expected_hex_prefix):
            hex_digits_after_prefix = 2 * (len(hex_response) - len(expected_hex_prefix))
            raise ValueError(
                f"Did not receive expected response:"
                f"Got '{hex_response}', expected {expected_hex_prefix}{'X' * hex_digits_after_prefix}"
            )

        return hex_response

    @staticmethod
    def __prepare_size(size: int) -> str:
        """Convert the size to its zero-padded 2-digit hex representation (256 is '00')"""
        if size <= 0 or size > 0xFF:
            raise ValueError(f"Size must be > 0 and <= 256, was {size}")
        return hex(size % 256)[2:].zfill(2)

    @staticmethod
    def __prepare_address(address: int) -> str:
        """Convert the address to its zero-padded 4-digit hex representation"""
        if address < 0 or address >= 0xFF:
            raise ValueError(f"Address must be >= 0 and < 256, was {address}")
        return hex(address)[2:].zfill(4)

    @staticmethod
    @lru_cache()
    def __compute_crc(command: bytes) -> int:
        """Compute the CRC checksum"""

        def _helper_func(c: int) -> int:
            crc = 0
            c <<= 8
            for _ in range(8):
                if (crc ^ c) & 0x8000:
                    crc = (crc << 1) ^ 0x1021
                else:
                    crc <<= 1
                c <<= 1
            return crc

        crc = 0
        for byte in command:
            crc = (crc << 8) ^ _helper_func(((crc >> 8) ^ (0xFF & byte)) & 0xFF)
            crc &= 0xFFFF

        return crc
